/*
  Copyright (C) JAEHYUK CHO
  All rights reserved.
  Code by JaeHyuk Cho <mailto:minzkn@minzkn.com>
*/

#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>

#include <pthread.h>

#define def_mztimer_realtime_signal SIGRTMIN
    
static long g_mztimer_interval_usec;

static void mztimer_handler(int s_signal, siginfo_t *s_siginfo, void *s_user_context)
{
    static unsigned int s_tick = 0;
    timer_t *s_timer_id_ptr;
    int s_overrun_count;

    (void)s_signal;
    (void)s_siginfo;
    (void)s_user_context;

    s_timer_id_ptr = (timer_t *)s_siginfo->si_value.sival_ptr;
    s_overrun_count = timer_getoverrun(*s_timer_id_ptr);
    if(s_overrun_count == (-1)) {
        s_overrun_count = 0;
    }

    (void)fprintf(stdout, "tick=%u, overrun_count=%d", s_tick, s_overrun_count);

#if 1L /* DEBUG: compare with system time */
    do {
        static unsigned long long s_start_time_usec = 0ull;
        unsigned long long s_current_time_usec;
        struct timeval s_timeval;
        unsigned long long s_duration_usec, s_diff_usec;

        s_duration_usec = s_tick * g_mztimer_interval_usec;

        (void)gettimeofday(&s_timeval, NULL);
        s_current_time_usec = (s_timeval.tv_sec * 1000000) + s_timeval.tv_usec;
        if(s_start_time_usec == 0ull) {
            s_start_time_usec = s_current_time_usec;
        }
        s_current_time_usec -= s_start_time_usec;

	if(s_duration_usec > s_current_time_usec) {
	    s_diff_usec = s_duration_usec - s_current_time_usec;
	}
	else {
	    s_diff_usec = s_current_time_usec - s_duration_usec;
	}

        fprintf(stdout, ", real=%llu.%03llu, diff=%llu.%llu", s_current_time_usec / 1000ull, s_current_time_usec % 1000ull, s_diff_usec / 1000ull, s_diff_usec % 1000ull);
    }while(0);
#endif

    (void)fprintf(stdout, "\n");

    s_tick += 1 + s_overrun_count;
}
  
static void *mztimer_thread(void *s_argument)
{
    static timer_t s_timer_id;
    struct sigevent s_sigevent;
    struct itimerspec s_itimerspec;
    sigset_t s_mask;
    struct sigaction s_sigaction;
    clockid_t s_clock_id = CLOCK_REALTIME;

    /* establish handler for timer signal */
    s_sigaction.sa_flags = SA_SIGINFO;
    s_sigaction.sa_sigaction = mztimer_handler;
    sigemptyset(&s_sigaction.sa_mask);
    if(sigaction(def_mztimer_realtime_signal, &s_sigaction, NULL) == (-1)) {
        return(NULL);
    }

    /* block timer signal temporarily */
    sigemptyset(&s_mask);
    sigaddset(&s_mask, def_mztimer_realtime_signal);
    if(sigprocmask(SIG_SETMASK, &s_mask, NULL) == (-1)) {
        return(NULL);
    }

    /* create timer */
    s_sigevent.sigev_notify = SIGEV_SIGNAL;
    s_sigevent.sigev_signo = def_mztimer_realtime_signal;
    s_sigevent.sigev_value.sival_ptr = &s_timer_id;
    if(timer_create(s_clock_id, &s_sigevent, &s_timer_id) == (-1)) {
        return(NULL);
    }

    /* start the timer */
    s_itimerspec.it_value.tv_sec = g_mztimer_interval_usec / 1000000l;
    s_itimerspec.it_value.tv_nsec = (g_mztimer_interval_usec % 1000000l) * 1000l;
    s_itimerspec.it_interval.tv_sec = s_itimerspec.it_value.tv_sec;
    s_itimerspec.it_interval.tv_nsec = s_itimerspec.it_value.tv_nsec;
    if(timer_settime(s_timer_id, 0, &s_itimerspec, NULL) == (-1)) {
        return(NULL);
    }
   
    /* unlock the timer signal */
    if(sigprocmask(SIG_UNBLOCK, &s_mask, NULL) == (-1)) {
        return(NULL);
    }

    for(;;) {
        pause();
    }

    return(NULL);
}

int mztimer_start(long s_interval_usec)
{
    pthread_t s_thread_handle;
    sigset_t s_mask;

    g_mztimer_interval_usec = s_interval_usec;

    if(pthread_create(&s_thread_handle, NULL, mztimer_thread, NULL) == (-1)) {
        return(-1);
    }
    (void)pthread_detach(s_thread_handle);
    
    sigemptyset(&s_mask);
    sigaddset(&s_mask, def_mztimer_realtime_signal);
    (void)sigprocmask(SIG_BLOCK, &s_mask, NULL);

    return(0);
}

int main(void)
{
    (void)mztimer_start(500l /* msec */ * 1000 /* usec */);

    for(;;) {
        
        sleep(1);
	
	(void)fprintf(stdout, "\x1b[1;33msleep loop\x1b[0m\n");
    }
    

    return(EXIT_SUCCESS);
}

/* End of source */
